import json
import logging
import sys
from copy import deepcopy as _copy
from hashlib import sha256
from tempfile import TemporaryDirectory

import cloudpickle
import numpy as np
import pandas as pd

import goldilox
from goldilox.config import AWS_PROFILE, PIPELINE_TYPE, VAEX, SKLEARN, BYTE_DELIMITER, VERSION, PY_VERSION, \
    PACKAGES, VARIABLES, DESCRIPTION, RAW
from goldilox.utils import _is_s3_url

logger = logging.getLogger()


class Pipeline:
    pipeline_type: str
    description: str
    BYTES_SIGNETURE = b"Goldilox"

    @classmethod
    def check_hash(cls, file_path):
        h = sha256()

        with open(file_path, "rb") as file:
            while True:
                # Reading is buffered, so we can read smaller chunks.
                chunk = file.read(h.block_size)
                if not chunk:
                    break
                h.update(chunk)

        return h.hexdigest()

    @staticmethod
    def to_raw(df):
        """
        Retrinve
        @param df: a Pandas Dataframe, Pandas Series, Vaex Dataframe, or numpy array.
        @return: a dict example of raw data which you would expect in production inference time.
        """
        if hasattr(df, "to_pandas_df"):  # vaex
            return df.to_records(0)
        elif isinstance(df, np.ndarray):  # numpy
            return list(df[0])
        elif isinstance(df, pd.Series):  # pandas Series
            return {df.name: df[0]}
        return df.iloc[0].to_dict()  # pandas

    @classmethod
    def from_vaex(cls, df, fit=None, variables=None, description="", validate=True):
        """
        Get a Pipeline out of a vaex.dataframe.DataFrame, and validate serilization and missing values.
        @param df: vaex.dataframe.DataFrame
        @param fit: method: A method which accepts a vaex dataframe and returns a vaex dataframe which run on pipeline.fit(df).
        @param variables: dict [optional]: Any variables we want to associate with the current pipeline.
        @param description: str [optional]: Any text we want to associate with the current pipeline.
        @param validate: bool [optional]: If true, run validation.
        @return: VaexPipeline
        """
        from goldilox.vaex.pipeline import VaexPipeline as VaexPipeline
        pipeline = VaexPipeline.from_dataframe(df=df, fit=fit, variables=variables, description=description)
        if validate:
            logger.info("validate pipeline")
            logger.info(f"pipeline valid: {pipeline.validate()}")
        return pipeline

    @staticmethod
    def _is_sklearn_fitted(pipeline):
        return hasattr(pipeline, "__sklearn_is_fitted__") and pipeline.__sklearn_is_fitted__()

    @classmethod
    def from_sklearn(
            cls,
            pipeline,
            raw=None,
            target=None,
            features=None,
            output_columns=None,
            variables=None,
            fit_params=None,
            description="",
            validate=True
    ):
        """
        :param sklearn.preprocessing.pipeline.Pipeline pipeline: The skleran pipeline
        :param raw: dict [optional]: An example of data which will be queried in production (only the features)
                - If X is provided, would be the first row.
        :param features: list [optional]: A list of columns - if X is provided, will take its columns - important if data provided as numpy array.
        :param target: str [optional]: The name of the target column - Used for retraining
        :param output_columns: List[str] [optional]: For sklearn output column in case the output is a numpy array
        :param variables: dict [optional]: Variables to associate with the pipeline - fit_params automatically are added
        :param description: str [optional]: A pipeline description and notes in text
        :param validate: bool [optional]: If True, run validation.
        :return: SkleranPipeline object
        """
        from goldilox.sklearn.pipeline import SklearnPipeline

        ret = SklearnPipeline.from_sklearn(
            pipeline=pipeline,
            features=features,
            target=target,
            raw=raw,
            output_columns=output_columns,
            variables=variables,
            fit_params=fit_params,
            description=description,
        )
        if validate and Pipeline._is_sklearn_fitted(pipeline):
            logger.info("validate pipeline")
            logger.info(f"pipeline valid: {ret.validate()}")
        return ret

    @classmethod
    def _read_file(cls, path):
        def open_state():
            return open(path, "rb")

        if _is_s3_url(path):
            import s3fs
            fs = s3fs.S3FileSystem(profile=AWS_PROFILE)

            def open_state():
                return fs.open(path, "rb")

        with open_state() as f:
            state_bytes = f.read()
        return Pipeline._split_meta(state_bytes)

    @classmethod
    def from_file(cls, path):
        """
        Read a pipeline from file.
        @param path: path to pipeline file.
        @return: SkleranPipeline or VaexPipeline.
        """
        meta_bytes, state_bytes = Pipeline._read_file(path)
        try:
            state = cloudpickle.loads(state_bytes)
            meta = cloudpickle.loads(meta_bytes)
        except:
            import pickle
            logger.warning("issue with cloudpickle loads")
            state = pickle.loads(state_bytes)
            meta = pickle.loads(meta_bytes)
        pipeline_type = meta.get(PIPELINE_TYPE)
        if pipeline_type == SKLEARN:
            return state
        elif pipeline_type == VAEX:
            from goldilox.vaex.pipeline import VaexPipeline
            return VaexPipeline.load_state(state)
        raise RuntimeError(f"Cannot load pipeline of type {pipeline_type} from {path}")

    @classmethod
    def load(cls, path):
        """Alias to from_file()"""
        return cls.from_file(path)

    @classmethod
    def load_meta(cls, path):
        """Read the meta information from a pipeline file without loading it"""
        meta_bytes, _ = Pipeline._read_file(path)
        return cloudpickle.loads(meta_bytes)

    def _get_meta(self):
        state = {
            PIPELINE_TYPE: self.pipeline_type,
            VERSION: goldilox.__version__,
            PY_VERSION: sys.version.split(" ")[0],
            PACKAGES: self._get_packages(),
            VARIABLES: self.variables.copy(),
            DESCRIPTION: self.description,
            RAW: self.raw,
        }
        return cloudpickle.dumps(_copy(state))

    @classmethod
    def _split_meta(cls, b):
        splited = b.split(BYTE_DELIMITER)
        return splited[0][len(Pipeline.BYTES_SIGNETURE):], splited[1]

    def save(self, path):
        """Save a pipeline to a file"""
        state_to_write = Pipeline.BYTES_SIGNETURE + self._get_meta() + BYTE_DELIMITER + self._dumps()
        if _is_s3_url(path):
            import s3fs
            fs = s3fs.S3FileSystem(profile=AWS_PROFILE)
            with fs.open(path, "wb") as outfile:
                outfile.write(state_to_write)
        else:
            try:
                import os

                if "/" in path:
                    os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
            except AttributeError as e:
                pass
            with open(path, "wb") as outfile:
                outfile.write(state_to_write)

        return path

    def validate(self, df=None, check_na=True, verbose=True):
        """
        Validate the pieline can be saved, reload, and run predictions.
        Can also check if missing value are handled.
        @param df: DataFrame [optional] - used to test prediction.
        @param check_na: bool [optional] - If true, test if missing data is handled.
        @param verbose: If True, log the validation.
        @return: True if pipeline can be served for predictions.
        """
        if verbose:
            logger.info("validate serialization")
        pipeline = Pipeline.from_file(self.save(TemporaryDirectory().name + "models/model.pkl"))
        if df is not None or self.raw is not None:
            X = df if df is not None else self.infer(self.raw)
            if verbose:
                logger.info("validate inference")
            assert len(pipeline.inference(X)) == len(X)
            if check_na:
                pipeline._validate_na()
        elif verbose:
            logger.info("No data, and no raw example existing for inference validation - skip")
        return True

    def _validate_na(self):
        ret = True
        copy = self.raw.copy()
        for column in copy:
            tmp = copy.copy()
            tmp[column] = None
            try:
                with np.errstate(all='ignore'):
                    self.inference(tmp).values
            except Exception as e:
                ret = False
                logger.warning(f"Pipeline doesn't handle NA for {column}")
        return ret

    def fit(self, df, **kwargs):
        return self

    def transform(self, df, **kwargs):
        raise NotImplementedError(f"Not implemented for {self.pipeline_type}")

    def predict(self, df, **kwargs):
        raise NotImplementedError(f"Not implemented for {self.pipeline_type}")

    def inference(self, df, **kwargs):
        raise NotImplementedError(f"Not implemented for {self.pipeline_type}")

    def infer(self, df):
        raise NotImplementedError(f"Not implemented for {self.pipeline_type}")

    @classmethod
    def to_records(cls, items):
        """Return data as records: [{key: value, ...}, ...]"""
        if isinstance(items, pd.DataFrame):
            return items.to_dict(orient="records")
        elif isinstance(items, list) or isinstance(items, dict):
            return items
        # vaex
        return items.to_records()

    @classmethod
    def jsonify(cls, items):
        """Return data as json: '[{key: value, ...}, ...]'"""
        if isinstance(items, pd.DataFrame):
            return items.to_json(orient="records")
        elif isinstance(items, list) or isinstance(items, dict):
            return json.dumps(items)
        # vaex
        return json.dumps(items.to_records())

    @staticmethod
    def _get_packages():
        """Run pip freeze and returns the results"""
        import subprocess
        return subprocess.check_output([sys.executable, '-m', 'pip',
                                        'freeze']).decode()



# Copyright (c) 2018 by contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8
import tempfile
import warnings

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_array, check_X_y
from xlearn import create_linear, create_fm, create_ffm
from xlearn.data import DMatrix
from copy import copy as _copy
import os

from threading import Lock
class PickableLock():
    def __init__(self):
        self.lock = Lock()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, type, value, traceback):
        self.lock.release()

    def __reduce__(self):
        return (self.__class__,())
lock = PickableLock()

def write_data_to_xlearn_format(X, y, filepath, fields=None):
    """ Write data to xlearn format (libsvm or libffm). Modified from
    https://github.com/scikit-learn/scikit-learn/blob/a24c8b46/sklearn/datasets/svmlight_format.py

    :param X: array-like
              Feature matrix in numpy or sparse format
    :param y: array-like
              Label in numpy or sparse format
    :param filepath: file location for writing data to
    :param fields: An array specifying fields in each columns of X. It should have same length
        as the number of columns in X. When set to None, convert data to libsvm format else
        libffm format.
    """

    with open(filepath, "wb") as f_handle:
        X_is_sp = int(hasattr(X, "tocsr"))
        y_is_sp = int(hasattr(y, "tocsr"))

        if X.dtype.kind == 'i':
            value_pattern = u"%d:%d"
        else:
            value_pattern = u"%d:%.16g"

        if fields is not None:
            is_ffm_format = True
            value_pattern = u"%d:" + value_pattern
        else:
            is_ffm_format = False

        if y.dtype.kind == 'i':
            label_pattern = u"%d"
        else:
            label_pattern = u"%.16g"

        line_pattern = u"%s %s\n"

        for i in range(X.shape[0]):
            if X_is_sp:
                span = slice(X.indptr[i], X.indptr[i + 1])
                x_indices = X.indices[span]
                row = zip(fields[x_indices], x_indices, X.data[span]) if is_ffm_format \
                    else zip(x_indices, X.data[span])
            else:
                nz = X[i] != 0
                row = zip(fields[nz], np.where(nz)[0], X[i, nz]) if is_ffm_format \
                    else zip(np.where(nz)[0], X[i, nz])

            if is_ffm_format:
                s = " ".join(value_pattern % (f, j, x) for f, j, x in row)
            else:
                s = " ".join(value_pattern % (j, x) for j, x in row)

            if y_is_sp:
                labels_str = label_pattern % y.data[i]
            else:
                labels_str = label_pattern % y[i]

            f_handle.write((line_pattern % (labels_str, s)).encode('ascii'))



class BaseXLearnModel(BaseEstimator):
    """ Implementation of Scikit-learn interface for xlearn models.

    :param model_type: one of 'lr', 'fm', 'ffm'
    :param task: 'binary' for classification or 'reg' for regression
    :param metric: 'acc', 'prec', 'recall', 'f1', 'auc' for classification,
        and 'mae', 'mape', 'rmsd (rmse)' for regression.
    :param log: location of log
    :param lr: learning rate
    :param k: latent factor for factorization
    :param reg_lambda: alias for lambda
    :param init: initial value
    :param fold: number of fold used in cross validation
    :param epoch: number of training epoch
    :param stop_window: window size for early stopping
    :param opt: optimizer option, one of 'sgd', 'adagrad', 'ftrl'
    :param nthread: number of threads (Deprecated, please use n_jobs)
    :param n_jobs: number of threads used to run xlearn.
    :param block_size: block size for on-disk training.
    :param alpha: alpha for FTRL
    :param beta: beta for FTRL
    :param lambda_1: lambda_1 for FTRL
    :param lambda_2: lambda_2 for FTRL
    :param kwargs: extra input arguments
    """

    def __init__(self, model_type='fm', task='binary', metric='auc', block_size=500,
                 lr=0.2, k=4, reg_lambda=0.1, init=0.1, fold=1, epoch=5, stop_window=2,
                 opt='sgd', nthread=None, n_jobs=4, alpha=1, beta=1, lambda_1=1, lambda_2=1,
                 seed=1, **kwargs):
        self.model_type = model_type
        self.task = task
        self.metric = metric
        self.lr = lr
        self.k = k
        self.reg_lambda = reg_lambda
        self.init = init
        self.fold = fold
        self.epoch = epoch
        self.opt = opt
        self.nthread = nthread
        self.n_jobs = n_jobs
        self.stop_window = stop_window
        self.alpha = alpha
        self.beta = beta
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.kwargs = kwargs
        self.block_size = block_size
        self.seed = seed
        self.lock = PickableLock()

        # initialize internal structure
        self._XLearnModel = None
        self._temp_model_file = tempfile.NamedTemporaryFile(delete=False)
        self._temp_weight_file = tempfile.NamedTemporaryFile(delete=False)
        self.weights = None
        self.fields = None

    def get_model(self):
        """ Return internal XLearn model.

        This will raise exception when model is not fitted

        :return: the underlying XLearn model
        """
        if self._XLearnModel is None:
            raise Exception('Need to call fit beforehand')

        return self._XLearnModel

    def get_params(self, deep=False):
        """ Get model parameters

        :param deep: is deep copy
        :return: model parameters
        """
        params = super(BaseXLearnModel, self).get_params(deep=deep)

        return params

    def get_xlearn_params(self):
        """ Get xlearn model parameters

        :return: model parameters used for training
        """
        xlearn_param = self.get_params()

        # rename reg_lambda as lambda, and remove model_type
        xlearn_param['lambda'] = xlearn_param.pop('reg_lambda')
        xlearn_param.pop('model_type')

        # rename n_jobs to nthread for _XLearnModel
        n_jobs = xlearn_param.pop('n_jobs')
        if 'nthread' in xlearn_param and xlearn_param['nthread'] is not None:
            warnings.warn("The nthread parameter has been deprecated. Please use n_jobs instead."
                          , DeprecationWarning)
        else:
            xlearn_param['nthread'] = n_jobs

        return xlearn_param

    def fit(self, X, y=None, fields=None,
            is_lock_free=True, is_instance_norm=True,
            eval_set=None, is_quiet=False):
        """ Fit the XLearn model given feature matrix X and label y

        :param X: array-like or a string specifying file location
                  Feature matrix
        :param y: array-like
                  Label
        :param fields: array-like
                  Fields for FFMModel. Default as None
        :param is_lock_free: is using lock-free training
        :param is_instance_norm: is using instance-wise normalization
        :param eval_set: a 2-element list representing (X_val, y_val) or a string specifying file location
        :param is_quiet: is training model quietly
        """
        if self._XLearnModel is None:
            if self.model_type == 'fm':
                self._XLearnModel = create_fm()
            elif self.model_type == 'lr':
                self._XLearnModel = create_linear()
            elif self.model_type == 'ffm':
                self._XLearnModel = create_ffm()
            else:
                raise Exception('model_type must be fm, ffm or lr')
        else:
            print('use old model')

        if y is None:
            assert isinstance(X, str), 'X must be a string specifying training file location' \
                                       ' when only X specified'
            self._XLearnModel.setTrain(X)

        else:
            X, y = check_X_y(X, y, accept_sparse=['csr'], y_numeric=True, multi_output=False)

            if self.model_type == 'ffm':
                assert fields is not None, 'Must specify fields in FFMModel'
                self.fields = fields

            # convert data into libsvm/libffm format for training
            train_set = DMatrix(X, y, self.fields)
            self._XLearnModel.setTrain(train_set)

        # TODO: find out what task need to set sigmoid
        if self.task == 'binary':
            self._XLearnModel.setSigmoid()

        # set lock-free, is quiet training and instance-wise normalization
        if not is_lock_free:
            self._XLearnModel.disableLockFree()

        if is_quiet:
            self._XLearnModel.setQuiet()

        if not is_instance_norm:
            if self.model_type in ['fm', 'ffm']:
                self._XLearnModel.disableNorm()
            else:
                warnings.warn('Setting is_instance_norm to False is ignored. It only applies to fm or ffm.')

        params = self.get_xlearn_params()

        # check if validation set exists or not
        if eval_set is not None:
            if isinstance(eval_set, str):
                self._XLearnModel.setValidate(eval_set)
            else:
                if not (isinstance(eval_set, list) and len(eval_set) == 2):
                    raise Exception('eval_set must be a 2-element list')

                # extract validation data
                X_val, y_val = check_X_y(eval_set[0], eval_set[1],
                                         accept_sparse=['csr'],
                                         y_numeric=True,
                                         multi_output=False)

                validate_set = DMatrix(X_val, y_val, self.fields)
                self._XLearnModel.setValidate(validate_set)

        # set up files for storing weights
        self._XLearnModel.setTXTModel(self._temp_weight_file.name)

        # fit model
        self._XLearnModel.fit(params, self._temp_model_file.name)

        # acquire weights
        self._parse_weight(self._temp_weight_file.name)

    def predict(self, X):
        """ Generate prediction using feature matrix X

        :param X: array-like
                  Feature matrix
        :return: prediction
        """

        if isinstance(X, str):
            self._XLearnModel.setTest(X)
        else:
            X = check_array(X, accept_sparse=['csr'])
            test_set = DMatrix(X, None, self.fields)
            self._XLearnModel.setTest(test_set)
        with self.lock:
            pred = self.get_model().predict(self._temp_model_file.name)
        return pred

    def feature_importance_(self):
        """TODO: analyze weight matrix to get feature importance"""
        pass

    def _convert_data(self, X, y, filepath, fields=None):
        """
        Convert feature matrix X and label y into libsvm format
        :param X: feature matrix
        :param y: label or None
        """
        if y is None:
            # create dummy label for test
            y = np.zeros(X.shape[0], dtype=np.int8)

        if fields is not None:
            # check if the model is ffm or not
            if self.model_type is not 'ffm':
                fields = None
                warnings.warn('Fields are ignored as it is not FFMModel')
            else:
                assert X.shape[1] == len(fields)

        try:
            write_data_to_xlearn_format(X, y, filepath, fields=fields)
        except:
            raise Exception('Failed to convert feature matrix X and label y to xlearn data format')

    def _parse_weight(self, file_name):
        # estimate number of features from txt file
        num_lines = sum(1 for line in open(file_name))
        num_features = int((num_lines - 1) / 2)

        if self.model_type in ['fm', 'ffm']:
            weight_vec = np.genfromtxt(file_name, skip_footer=num_features)
            weight_mtx = np.genfromtxt(file_name, skip_header=num_features + 1)
        else:
            weight_vec = np.genfromtxt(file_name)
            weight_mtx = None

        self.weights = (weight_vec, weight_mtx)

    def _remove_temp_file(self, temp_file):
        # The temp_file might be converted to binary file during training/inference.
        # remove both original temp_file and derived binary file if exist
        temp_bin_file = temp_file.name + '.bin'
        if os.path.exists(temp_bin_file):
            os.remove(temp_bin_file)
        temp_file.close()
        if os.path.exists(temp_file.name):
            os.remove(temp_file.name)

    def __del__(self):
        self._remove_temp_file(self._temp_model_file)
        self._remove_temp_file(self._temp_weight_file)

    def __delete__(self, instance):
        # self._temp_model_file.close()
        # self._temp_weight_file.close()
        del instance

    def __getstate__(self):
        state = self.__dict__
        state['_XLearnModel'] = None
        state = _copy(state)
        state['_temp_model_file'] = state['_temp_model_file'].read()
        state['_temp_weight_file'] = state['_temp_weight_file'].read()
        self.init_model()
        return state

    def init_model(self):
        if self._XLearnModel is None:
            if self.model_type == 'fm':
                self._XLearnModel = create_fm()
            elif self.model_type == 'lr':
                self._XLearnModel = create_linear()
            elif self.model_type == 'ffm':
                self._XLearnModel = create_ffm()
        if self._temp_model_file:
            self._XLearnModel.setPreModel(self._temp_model_file.name)

    def __setstate__(self, state):
        self.__dict__ = state
        _temp_model_file = tempfile.NamedTemporaryFile(delete=False)
        if self._temp_model_file is not None:
            with open(_temp_model_file.name, 'wb') as f:
                f.write(self._temp_model_file)
            self._temp_model_file = _temp_model_file
        _temp_weight_file = tempfile.NamedTemporaryFile(delete=True)
        if self._temp_weight_file is not None:
            with open(_temp_weight_file.name, 'wb') as f:
                f.write(self._temp_weight_file)
            self._temp_weight_file = _temp_weight_file
        self.init_model()


class FMModel(BaseXLearnModel):
    """ Factorization machine (FM) model
    """

    def __init__(self, model_type='fm', task='binary', metric='auc', block_size=500,
                 lr=0.2, k=4, reg_lambda=0.1, init=0.1, fold=1, epoch=5, stop_window=2,
                 opt='sgd', nthread=None, n_jobs=4, alpha=1, beta=1, lambda_1=1, lambda_2=1,
                 seed=1, **kwargs):
        assert model_type == 'fm', 'Model type must be set to fm for FMModel'
        super(FMModel, self).__init__(model_type, task, metric, block_size,
                                      lr, k, reg_lambda, init, fold, epoch, stop_window,
                                      opt, nthread, n_jobs, alpha, beta, lambda_1, lambda_2,
                                      seed, **kwargs)

    def __delete__(self, instance):
        super(FMModel, self).__delete__(instance)


class LRModel(BaseXLearnModel):
    """ linear model
    """

    def __init__(self, model_type='lr', task='binary', metric='auc', block_size=500,
                 lr=0.2, k=4, reg_lambda=0.1, init=0.1, fold=1, epoch=5, stop_window=2,
                 opt='sgd', nthread=None, n_jobs=4, alpha=1, beta=1, lambda_1=1, lambda_2=1,
                 seed=1, **kwargs):
        assert model_type == 'lr', 'Model type must be set to lr for LRModel'
        super(LRModel, self).__init__(model_type, task, metric, block_size,
                                      lr, k, reg_lambda, init, fold, epoch, stop_window,
                                      opt, nthread, n_jobs, alpha, beta, lambda_1, lambda_2,
                                      seed, **kwargs)

    def __delete__(self, instance):
        super(LRModel, self).__delete__(instance)


class FFMModel(BaseXLearnModel):
    """ Field-aware factorization machine (FFM) model
    """

    def __init__(self, model_type='ffm', task='binary', metric='auc', block_size=500,
                 lr=0.2, k=4, reg_lambda=0.1, init=0.1, fold=1, epoch=5, stop_window=2,
                 opt='sgd', nthread=None, n_jobs=4, alpha=1, beta=1, lambda_1=1, lambda_2=1,
                 seed=1, **kwargs):
        assert model_type == 'ffm', 'Model type must be set to ffm for FFMModel'
        super(FFMModel, self).__init__(model_type, task, metric, block_size,
                                       lr, k, reg_lambda, init, fold, epoch, stop_window,
                                       opt, nthread, n_jobs, alpha, beta, lambda_1, lambda_2,
                                       seed, **kwargs)

    def __delete__(self, instance):
        super(FFMModel, self).__delete__(instance)



import json
import shutil
import subprocess
import sys
from pathlib import Path

import click

import goldilox
from goldilox import Pipeline
from goldilox.config import VARIABLES, RAW, DESCRIPTION, PACKAGES
from goldilox.utils import process_variables


@click.group()
@click.version_option("1.0.0")
def main():
    """This runs every time"""
    pass


def process_option(s):
    splited = s.split('=')
    if len(splited) != 2:
        splited = s.split(' ')
    if len(splited) != 2:
        return None, None
    key, value = splited[0], splited[1]
    if key.startswith('--'):
        key = key[2:]
    return key, value


@main.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.argument("path", type=click.Path(exists=True))
@click.argument('options', nargs=-1, type=click.UNPROCESSED)
def serve(path, **options):
    """Serve a  pipeline with fastapi server"""
    server_options = {}
    for option in options['options']:
        key, value = process_option(option)
        if key is None:
            print(f"option {option} was is illegal and was ignored ")
        else:
            server_options[key] = value
    from goldilox.app import Server
    Server(path, options=server_options).serve()


@main.command()
def arguments():
    """gunicorn arguments reference"""
    print("check https://docs.gunicorn.org/en/stable/run.html for options")


@main.command()
@click.argument("path", type=click.Path(exists=True))
def description(path):
    """print pipeline description"""
    meta = Pipeline.load_meta(path)
    click.echo(meta[DESCRIPTION])


@main.command()
@click.argument("path", type=click.Path(exists=True))
def example(path):
    """print pipeline output example with all possible outputs"""
    click.echo(json.dumps(Pipeline.load(path).example, indent=4))


@main.command()
@click.argument("path", type=click.Path(exists=True))
def raw(path):
    """print pipeline input example (raw data)"""
    meta = Pipeline.load_meta(path)
    click.echo(json.dumps(meta[RAW], indent=4))


@main.command()
@click.argument("path", type=click.Path(exists=True))
def variables(path):
    """print pipeline variables"""
    meta = Pipeline.load_meta(path)
    click.echo(json.dumps(process_variables(meta[VARIABLES]), indent=4))


@main.command()
@click.argument("path", type=click.Path(exists=True))
def packages(path):
    """print pipeline packages"""
    meta = Pipeline.load_meta(path)
    click.echo(json.dumps(meta[PACKAGES], indent=4))


def _write_content(output, content):
    with open(output, 'w') as outfile:
        outfile.write(content)
    return output


@main.command()
@click.argument("path", type=click.Path(exists=True))
@click.argument("output", type=str)
def freeze(path, output='requirements.txt'):
    """write pipeline packages to a file (pip freeze > output)"""
    packages = Pipeline.load_meta(path)[PACKAGES]
    _write_content(output, packages)
    click.echo(f"checkout {output} for the requirements")


@main.command()
@click.argument("path", type=click.Path(exists=True))
@click.option("--name", type=str, default="goldilox")
@click.option('--image', type=str, default=None)
@click.option('--platform', type=str, default=None)
def build(path, name="goldilox", image=None, platform=None):
    """ build a docker server image"""
    goldilox_path = Path(goldilox.__file__)
    docker_file_path = str(goldilox_path.parent.absolute().joinpath('app').joinpath('Dockerfile'))
    run_args = ['docker', 'build', f"-f={docker_file_path}", f"-t={name}"]
    build_args = ["--build-arg", f"PIPELINE_FILE={path}"]
    suffix_arg = ['.']
    if image is not None:
        build_args = build_args + ["--build-arg", f"PYTHON_IMAGE={image}"]
    if platform is not None:
        run_args = run_args + [f"--platform={platform}"]
    command = run_args + build_args + suffix_arg
    click.echo(f"Running docker build as follow:")
    click.echo(f"{' '.join(command)}")
    click.echo(f" ")
    subprocess.check_call(command)
    run_command = f"docker run --rm -it {name}" if platform is None else f"docker run --rm -it --platform={platform} -p 127.0.0.1:5000:8000 {name}"
    click.echo(f"Image {name} created - run with: '{run_command}'")


@main.command()
@click.option('--output', type=str, default=None)
def dockerfile(output):
    """Create a Dockerfile for you to work with"""
    goldilox_path = Path(goldilox.__file__)
    docker_file_path = str(goldilox_path.parent.absolute().joinpath('app').joinpath('Dockerfile'))
    docker_output = output or './Dockerfile'
    shutil.copyfile(docker_file_path, docker_output)
    with open(docker_output, 'r') as f:
        click.echo(f.read())
    click.echo("##################\nDockerfile was writen to './dockerfile'\n")
    click.echo(
        "Use 'docker build -f=Dockerfile -t=<image-name> --build-arg PIPELINE_FILE=<pipeline-path> --build-arg PYTHON_IMAGE=<base-image> .' to build a docker image")


if __name__ == '__main__':
    args = sys.argv
    if "--help" in args or len(args) == 1:
        print("Help stuff")
    main()



import logging
import re
from uuid import uuid4

import vaex
from traitlets import traitlets
from vaex.ml.state import HasState

logger = logging.getLogger()


class TransformerBase(HasState):
    output_columns = traitlets.List(traitlets.Unicode(), allow_none=True, help='List of the columns which were set')
    features = traitlets.List(traitlets.Unicode(), allow_none=True, help='List of the features as input')

    def fit(self, df, **kwargs):
        super().fit(df)
        return self

    def transform(self, df, **kwargs):
        return super().transform(df)

    def fit_transform(self, df, **kwargs):
        self.fit(df, **kwargs)
        return self.transform(df, **kwargs)

    @classmethod
    def load_state(cls, state):
        instance = cls()
        instance.state_set(state)
        return instance

    def preprocess_transform(self, df):
        copy = df.copy()
        return copy


@vaex.serialize.register
class Imputer(TransformerBase):
    strategy = traitlets.Dict(help='Strategy for handling missing values')
    prefix = traitlets.Unicode(default_value='', help="")
    values = traitlets.Dict(help='The value to fill in each feature')
    warnings = traitlets.Bool(default_value=True, help="Raise warnings when can not fillna a column")
    MEAN = 'MEAN'
    MODE = 'MODE'
    NEW_VALUE = 'NEW_VALUE'
    VALUE = 'VALUE'
    COMMON = 'COMMON'

    @staticmethod
    def mean(ar):
        return ar.mean()

    @staticmethod
    def mode(ar):
        return ar.df.percentile_approx(ar)

    @staticmethod
    def common(ar):
        return ar.value_counts(dropna=True).index[0]

    @staticmethod
    def new_value(ar):
        if ar.dtype.is_string:
            return f"NA_{str(uuid4())}"
        elif ar.dtype.is_numeric:
            return int(ar.max()) + 1
        raise ValueError(f"Can not create a new value for {ar.expression} of {ar.dtype} dtype")

    def get_value(self, ar):
        # TODO run in one go
        dtype = ar.dtype
        name = ar.expression
        feature_strategy = None

        try:
            # TODO fix whe NotImplementedError: large_string is fixed
            dtype_name = dtype.name
        except:
            dtype_name = None
        try:
            shape = ar.shape
        except:
            shape = (1,)
        if name in self.strategy:  # by name
            feature_strategy = self.strategy.get(name)
        elif 1 < len(shape):  # nd arrays
            feature_strategy = None
        # by dtype
        elif dtype_name is not None and dtype_name in self.strategy:
            feature_strategy = self.strategy.get(dtype_name.name)
        elif dtype.is_integer and int in self.strategy:
            feature_strategy = self.strategy.get(int)
        elif dtype.is_float and float in self.strategy:
            feature_strategy = self.strategy.get(float)
        elif dtype == bool and bool in self.strategy:
            feature_strategy = self.strategy.get(bool)
        elif dtype == str and str in self.strategy:
            feature_strategy = self.strategy.get(str)
        if feature_strategy is None and len(shape) == 1:
            feature_strategy = self.get_default_value(dtype)
        if callable(feature_strategy):
            feature_strategy = feature_strategy(ar)

        value = feature_strategy
        if dtype.is_string:
            if not isinstance(value, str):
                value = str(value)
            if feature_strategy == self.COMMON:
                value = self.common(ar)
            elif feature_strategy == self.NEW_VALUE:
                value = self.new_value(ar)
        elif dtype.is_numeric and len(shape) == 1:
            if isinstance(feature_strategy, str):
                if feature_strategy == self.MEAN:
                    value = float(ar.mean())
                elif feature_strategy == self.MODE:
                    value = float(self.mode(ar))
                elif feature_strategy == self.COMMON:
                    value = float(self.common(ar))
                elif feature_strategy == self.NEW_VALUE:
                    value = float(self.new_value(ar))

            if not isinstance(value, (int, float)):
                raise RuntimeError(f"value {value} cannot be used for {name} of type {dtype}")

        if value is None and self.warnings:
            logger.warning(f"value {value} was 'None' for {name} of type {dtype}")
            # raise RuntimeError(f"value {value} was 'None' for {name} of type {dtype}")
        return value

    def preprocess_fit(self, df, **kwargs):
        self.set_features(df)
        return df

    def set_features(self, df):
        if self.features is None or len(self.features) == 0:
            columns = df.get_column_names()
            self.features = [column for column in columns if
                             df[column].dtype.is_primitive or df[column].dtype.is_string]
        return self.features

    def fit(self, df, **kwargs):
        copy = df.copy()
        self.preprocess_fit(df)
        self.output_columns = []
        for feature in self.features:
            self.values[feature] = self.get_value(copy[feature])
            if self.values.get(feature) is not None and feature in self.features:
                name = self.prefix + feature
                self.output_columns.append(name)
        return self

    def add_column(self, df, name, value, first_column):
        df[name] = df.func.where(df[first_column] == value, value, value)
        return df

    def preprocess_transform(self, df):
        copy = df.copy()
        first_column = copy.get_column_names()[0]
        for column, value in self.values.items():
            if column not in copy:
                copy = self.add_column(copy, str(column), value, first_column)
        return copy

    def transform(self, df, **kwargs):
        copy = self.preprocess_transform(df)
        for column, value in self.values.items():
            name = self.prefix + column
            if value is None:
                if self.warnings:
                    logger.warning(f" could not fillna for {name} ")
                continue
            copy[name] = copy[column].fillna(value)
        return copy

    def get_default_value(self, dtype):
        if dtype.is_string:
            return ''
        if dtype.is_numeric:
            return 0
        try:
            if dtype.name == 'bool':
                return False
        except:
            pass
        if dtype == object:
            return {}
        if self.warnings:
            logger.warning(f"dtype {dtype} has no default value")
        return None

        # raise ValueError(f"dtype {dtype} has no default value")

    def __repr__(self):
        return f"Imputer({self.strategy})"

    @staticmethod
    def _to_type(s):
        return eval(re.sub('[^0-9a-zA-Z]+', '', s.replace('class', '')))

    @staticmethod
    def _is_type(s):
        return isinstance(s, str) and s.startswith('<class')

    def state_get(self):
        state = TransformerBase.state_get(self)
        for key, value in state['strategy'].items():
            if isinstance(key, type):
                state['strategy'][str(key)] = state['strategy'].pop(key)
        return state

    def state_set(self, state, trusted=True):
        if 'strategy' in state:
            for key, value in state['strategy'].items():
                if self._is_type(key):
                    state['strategy'][self._to_type(key)] = state['strategy'].pop(key)
        TransformerBase.state_set(self, state)



# import pickle
# from base64 import b64encode
#
# registry = {}
#
#
# def register(cls):
#     registry[fullname(cls)] = cls
#     return cls
#
#
# def fullname(cls):
#     return cls.__module__ + '.' + cls.__name__
#
#
# def to_dict(obj):
#     return dict(cls=fullname(obj.__class__), state=obj.state_get())
#
#
# def from_dict(d, trusted=False):
#     cls_name = d['cls']
#     if cls_name not in registry:
#         # lets load the module, so we give it a chance to register
#         module, dot, cls = cls_name.rpartition('.')
#         __import__(module)
#     if cls_name not in registry:
#         raise ValueError('unknown class: ' + cls_name)
#     else:
#         obj = registry[cls_name].state_from(d['state'], trusted=trusted)
#         # obj.state_set(d['state'])
#         return obj
#
#
# def f(test=None):
#     return 1
#
#
# def encode_function(f):
#     return b64encode(pickle.dumps(f)).decode('ascii')
#
#
# def decode_function(encoding):
#     # blob = encoding.get_blob(spec['blob'])
#
#     return pickle.loads(blob)
#
#     return encoding.decode('pickle', spec['func'], trusted=trusted),
#
#
# encoding = b64encode(pickle.dumps(f)).decode('ascii')



from goldilox.app import Server


def main():
    path = 'tests/models/server.pkl'
    options = {'workers': 2,
               'bind': f"localhost:5000",
               'preload': True,
               'worker_class': "uvicorn.workers.UvicornH11Worker"
               }
    # can only create the app after starting redis

    Server(path, options).serve()


if __name__ == "__main__":
    main()



from io import StringIO, BytesIO
import collections
import numpy as np
import matplotlib.colors
import warnings
from base64 import b64encode

import vaex
import vaex.utils
from vaex.image import rgba_2_pil

PIL = vaex.utils.optional_import("PIL.Image", package_name="pillow")


@vaex.register_function(scope='image')
def open(paths):
    images = [PIL.Image.open(path) for path in vaex.array_types.tolist(paths)]
    return np.array(images, dtype="O")


@vaex.register_function(scope='image')
def as_numpy(images):
    images = [np.array(image) for image in images]
    return np.array(images)


@vaex.register_function(scope='image')
def resize(images, size, resample=3):
    images = [image.resize(size, resample=resample) for image in images]
    return np.array(images, dtype="O")


@vaex.register_function(scope='image')
def as_image(arrays):
    images = [rgba_2_pil(image_array) for image_array in arrays]
    return np.array(images, dtype="O")

def rgba_2_pil(rgba):
    with warnings.catch_warnings():
        pass


import logging
from time import time

import cloudpickle
import numpy as np
import pandas as pd
import traitlets

from goldilox import Pipeline
from goldilox.config import STATE

DEFAULT_OUTPUT_COLUMN = "prediction"
TRAITS = "_trait_values"

logger = logging.getLogger()


class SklearnPipeline(traitlets.HasTraits, Pipeline):
    pipeline_type = "sklearn"
    current_time = int(time())
    created = traitlets.Int(
        default_value=current_time, allow_none=False, help="Created time"
    )
    updated = traitlets.Int(
        default_value=current_time, allow_none=False, help="Updated time"
    )
    raw = traitlets.Any(
        default_value=None,
        allow_none=True,
        help="An example of the transformed dataset",
    )
    pipeline = traitlets.Any(allow_none=False, help="A sklearn pipeline")
    features = traitlets.List(allow_none=True, help="A list of features")
    target = traitlets.Unicode(allow_none=True, help="A column to learn on fit")
    output_columns = traitlets.List(allow_none=True, help="The output column names")
    fit_params = traitlets.Dict(
        allow_none=True, default_value={}, help="params to use on fit time"
    )
    description = traitlets.Unicode(allow_none=True,
                                    default_value="", help="Any notes to associate with a pipeline instance"
                                    )
    variables = traitlets.Dict(
        default_value={}, help="Any variables to associate with a pipeline instance"
    )

    @property
    def example(self):
        """Returns an example of all possible outputs"""
        return self.inference(self.raw).to_dict(orient="records")[0]

    def set_variable(self, key, value):
        self.variables[key] = value
        return value

    def get_variable(self, key, default=None):
        return self.variables.get(key, default)

    @classmethod
    def from_sklearn(
            cls,
            pipeline,
            raw=None,
            features=None,
            target=None,
            output_columns=None,
            variables=None,
            fit_params=None,
            description="",
    ):
        """
        :param sklearn.preprocessing.pipeline.Pipeline pipeline: The skleran pipeline
        :param raw: dict [optional]: An example of data which will be queried in production (only the features)
                - If X is provided, would be the first row.
        :param features: list [optional]: A list of columns - if X is provided, will take it's columns
        :param target: str [optional]: The name of the target column - Used for retraining
        :param output_columns: List[str] [optional]: For sklearn estimator which predict a numpy.ndarray, name the output columns.
        :param variables: dict [optional]: Variables to associate with the pipeline - fit_params automatically are added
        :param description: str [optional]: A pipeline description and notes in text
        :return: SkleranPipeline object
        """
        if isinstance(features, pd.core.indexes.base.Index):
            features = list(features)
        elif isinstance(raw, dict) and features is None:
            features = list(raw.keys())
        elif raw is None and isinstance(features, list):
            raw = {key: 0 for key in features}
        if (
                hasattr(pipeline, "__sklearn_is_fitted__")
                and pipeline.__sklearn_is_fitted__()
                and raw is None
                and features is None
        ):
            raise RuntimeError(
                "For a fitted pipeline, please provide either the 'features' or 'sample'"
            )
        if variables is None:
            variables = {}
        if fit_params:
            variables.update(fit_params)
        if hasattr(pipeline, 'predict') and not hasattr(pipeline, 'transform') and (
                output_columns is None or len(output_columns) == 0):
            output_columns = [DEFAULT_OUTPUT_COLUMN]
        return SklearnPipeline(
            pipeline=pipeline,
            features=features,
            target=target,
            raw=raw,
            output_columns=output_columns,
            fit_params=fit_params,
            variables=variables,
            description=description,
        )

    @property
    def fitted(self):
        """Returns True is the pipeline/model is fitted"""
        return self.pipeline is not None and self.pipeline.__sklearn_is_fitted__()

    def infer(self, df):
        """Turn many inputs into a dataframes"""
        if isinstance(df, pd.DataFrame):
            return df.copy()
        if isinstance(df, pd.Series):
            return pd.DataFrame({df.name: df})
        if isinstance(df, np.ndarray):
            if self.features and len(self.features) == df.shape[1]:
                return pd.DataFrame(df, columns=self.features)
            elif self.output_columns and len(self.output_columns) == df.shape[1]:
                return pd.DataFrame(df, columns=self.output_columns)
            else:
                return pd.DataFrame(df)
        if isinstance(df, dict):
            df = [df.copy()]
        if isinstance(df, list):
            return pd.DataFrame(df, columns=self.features)
        try:
            import vaex

            if isinstance(df, vaex.dataframe.DataFrame):
                return df.to_pandas_df()
        except:
            pass
        raise RuntimeError(f"could not infer type:{type(df)}")

    def _dumps(self):
        return cloudpickle.dumps(self)

    @classmethod
    def loads(cls, state):
        if isinstance(state, bytes):
            state = cloudpickle.loads(state)
        return state[STATE]

    @classmethod
    def from_file(cls, path):
        return Pipeline.from_file(path)

    def _to_pandas(self, X, y=None):
        try:
            import vaex
            if isinstance(X, vaex.dataframe.DataFrame):
                X = X.to_pandas_df()
            elif isinstance(X, pd.Series):
                X = pd.DataFrame({X.name: X})
            elif isinstance(X, vaex.expression.Expression):
                name = X.expression
                X = X.to_pandas_series()
                X.name = name
            if isinstance(y, vaex.expression.Expression):
                name = y.expression
                y = y.to_pandas_series()
                y.name = name
                self.target = name
        except:
            pass
        if isinstance(X, np.ndarray):
            X = self.infer(X)
        if y is None:
            y = X[self.target] if self.target else None
        elif isinstance(y, pd.Series):
            self.target = y.name
        elif isinstance(y, str):
            self.target = y
            y = X[y]
        X = self._set_features(X)
        return X, y

    def _set_features(self, X):
        if isinstance(X, pd.DataFrame):
            if self.features is None:
                self.features = list(X.columns)
        if self.features:
            if self.target in self.features:
                self.features.remove(self.target)
            if len(self.features) > 1:
                X = X[self.features]
            elif len(self.features) == 1:
                X = X[self.features[0]]
        return X

    def fit(self, df, y=None, validate=True, check_na=True):
        X, y = self._to_pandas(df, y)
        self.raw = self.to_raw(X)
        params = self.fit_params or {}
        self.pipeline = self.pipeline.fit(X=X, y=y, **params)
        if validate:
            self.validate(check_na=check_na)
        return self

    def transform(self, df, **kwargs):
        copy = self.infer(df)
        features = self.features or copy.columns
        copy = self.pipeline.transform(copy[features])
        if isinstance(copy, np.ndarray) and copy.shape[1] == len(features):
            copy = pd.DataFrame(copy, columns=features)
        return copy

    def inference(self, df, columns=None, passthrough=True, **kwargs):
        """
        Returns the transformed data.
        Always tries to return a dataframe.
        If self.pipeline implements predict, returns the predictions as a new column (self.output_columns[0]).
        Else, returns self.pipeline.transform(df)
        @param df:
        @param columns: Only this columns are returns from the data frame.
        @param passthrough: If True, extra columns in the data passthrough.
        @param kwargs:
        @return:
        """
        copy = self.infer(df)
        features = self.features or copy.columns
        if features is None:
            raise RuntimeError("Model is not trained yet")
        passthrough_data = None
        if len(features) == 1:  # for text transformers and likewise
            features = features[0]
        X = copy[features] if features else copy
        if hasattr(self.pipeline, "predict") and self.output_columns is not None and len(self.output_columns) == 1:
            copy[self.output_columns[0]] = self.pipeline.predict(X)
        else:
            if passthrough:
                passthrough_columns = [column for column in copy.columns if column not in features]
                if columns is not None and len(columns) > 0:
                    passthrough_columns = [column for column in passthrough_columns if column in columns]
                if len(passthrough_columns) > 0:
                    passthrough_data = copy[passthrough_columns]
            copy = self.pipeline.transform(X)
            if isinstance(copy, np.ndarray) and self.output_columns and copy.shape[1] == len(self.output_columns):
                copy = pd.DataFrame(copy, columns=self.output_columns)
        if passthrough_data is not None and passthrough_data.shape[0] == copy.shape[0]:
            if isinstance(copy, pd.DataFrame):
                for column in passthrough_columns:
                    copy[column] = passthrough_data[column].values
            elif isinstance(copy, np.ndarray):
                copy = np.column_stack((copy, passthrough_data.values))
        if isinstance(copy, pd.DataFrame) and columns is not None and len(columns) > 0:
            copy = copy[[column for column in columns if column in copy]]
        return copy



import gc
import logging
import os
from pathlib import Path

import numpy as np
import pandas as pd

from goldilox.config import DEFAULT_SUFFIX

valida_types = {type(None), dict, list, int, float, str, bool}


def _is_s3_url(path):
    if hasattr(path, 'dirname'):
        path = path.dirname
    return path.startswith('s3://')


logger = logging.getLogger()


# s3_client = boto3.client('s3')


def open_many(paths):
    import vaex
    dfs = [vaex.open(path) for path in paths]
    valid_eips = set()
    info_columns = set()
    for df in dfs:
        eips = df.get_column_names(regex='eip')
        for column in eips:
            if len(df[column].as_numpy().unique()) > 1:
                valid_eips.add(column)
        info_columns.update(df.get_column_names(regex='^(?!eip_).+'))
    columns = list(info_columns) + list(valid_eips)

    for column in columns:
        for df, p in zip(dfs, paths):
            if column not in df:
                logger.info(f"added missing column {column} to {p}")
                df[column] = np.full(shape=len(df), fill_value=None)
    pandas = [df[columns].to_pandas_df() for df in dfs]
    for df in pandas:
        for column in valid_eips:
            df[column] = df[column].astype(str)
    concat = pd.concat(pandas)
    concat_vaex = vaex.from_pandas(concat)
    del concat
    del dfs
    del pandas
    gc.collect()
    return concat_vaex


def read_data(path, prefix=None, suffix=DEFAULT_SUFFIX):
    import vaex
    prefix = prefix or ''
    logger.info(f"read data from {path} and prefix {prefix} and suffix {suffix}")
    if os.path.isdir(path):
        files = [str(path) for path in Path(path).rglob(f"{prefix}*{suffix}")]
        logger.info(f"relevant files: {files}")
        logger.info(f"found {len(files)} files")
        if len(files) == 0:
            logger.error(f"found no data files")
            return None

        bad_files = []
        for file in files:
            try:
                if vaex.open(file).head(2):
                    pass
            except:
                logger.error(f"file {file} is broken")
                bad_files.append(file)

        paths = list(set([file for file in files if file not in bad_files]))
        logger.info(f"found {len(bad_files)} malformed files")
        logger.info(f"open {len(paths)} files")
        df = vaex.open_many(paths)
    else:
        df = vaex.open(path, shuffle=True)
    logger.info(f"data shape {df.shape}")
    return df


def is_list(l):
    return isinstance(l, (list, pd.core.series.Series, np.ndarray))


def to_nulls(value):
    if is_list(value):
        return [to_nulls(v) for v in value]
    elif isinstance(value, dict):
        return {to_nulls(k): to_nulls(v) for k, v in value.items()}
    elif hasattr(value, 'tolist'):
        return to_nulls(value.tolist())
    elif pd.isnull(value):
        return None
    return value


def process_variables(variables):
    return {
        key: to_nulls(value)
        for key, value in variables.items()
        if (type(value) in valida_types or hasattr(value, 'tolist'))
    }



import logging

import numpy as np
import vaex

logger = logging.getLogger(name='VowpalWabbit')

CATEGORICAL = 'categorical'
TEXT = 'text'

@vaex.register_dataframe_accessor('vw', override=True)
class DataFrameAccessorTensorflow(object):
    def __init__(self, df):
        self.df = df
        self.column_map = None
        self.numeric_features_indices = None
        self.categotical_features_indices = None
        self.text_features_indices = None
        self.target_index = None
        self.weight_index = None
        self.row_length = None

    def init_generator(self, features, target=None, weights=None, categorical_features=[], text_features=[]):
        target = target
        numeric_features_indices = set([])
        categotical_features_indices = set([])
        text_features_indices = set([])
        self.target_index = None
        self.weight_index = None
        non_numeric = categorical_features + text_features
        self.column_map = {i: name for i, name in enumerate(self.df.get_column_names())}
        all_columns = self.df.get_column_names()
        for i, column in enumerate(all_columns):
            if column == target:
                self.target_index = i
            elif column == weights:
                self.weight_index = i
            elif column not in non_numeric and self.df[column].dtype.is_numeric:
                numeric_features_indices.add(i)
            elif column in categorical_features:
                categotical_features_indices.add(i)
            elif column in text_features:
                text_features_indices.add(i)
            elif self.df[column].dtype.is_string:
                categotical_features_indices.add(i)
        assert len(categotical_features_indices) + len(text_features_indices) + len(numeric_features_indices) == len(
            features)
        self.numeric_features_indices = numeric_features_indices
        self.categotical_features_indices = categotical_features_indices
        self.text_features_indices = text_features_indices
        self.row_length = len(features)
        if target is not None:
            self.row_length += 1
        if weights is not None:
            self.row_length += 1


    def to_vw_generator(self, features, target=None, weights=None, categorical_features=[], text_features=[], epochs=5,
                        verbose=True):
        self.init_generator(features=features,target=target,weights=weights,categorical_features=categorical_features,text_features=text_features)

        def _generator(verbose=verbose):
            length = len(self.df)
            for epoch in range(epochs):
                if verbose:
                    logger.info(f"start epoch {epoch}")
                for i in range(length):
                    row = self.df[i]
                    ex = self.to_vw(row)

                    if verbose and i == 0:
                        logger.info(f"example first row {i}: {ex}")
                    yield ex

        return _generator()

    def to_vw(self, row):
        row = list(row)
        weight = ''
        truth = ''
        if len(row) < self.row_length:
            if self.target_index is not None:
                row.insert(self.target_index, truth)
            if self.weight_index is not None:
                row.insert(self.weight_index, weight)
        numeric_strings = []
        categorical_strings = []
        text_strings = []
        for index, value in enumerate(row):
            if value is None:
                continue
            name = self.column_map.get(index)
            if index == self.target_index:
                truth = value
            elif index == self.weight_index:
                weight = value
            elif index in self.numeric_features_indices:
                if np.isnan(value):
                    continue
                numeric_strings.append(f"{name}:{value}")
            elif index in self.categotical_features_indices:
                categorical_strings.append(f"{name}_{str(value).replace(' ', '_')}")
            elif index in self.text_features_indices:
                text_strings.append(f"{value}")
        values = [f"{truth} {weight}"]
        if 0 < len(numeric_strings) > 0:
            values.append(' '.join(numeric_strings))
        if 0 < len(categorical_strings):
            values.append(' '.join([CATEGORICAL] + categorical_strings))
        if 0 < len(text_strings):
            values.append(' '.join([TEXT]+ text_strings))

        ex = '|'.join(values)
        return ex



import os
from distutils.util import strtobool

AWS_PROFILE = os.environ.get("AWS_PROFILE")

# Constants
STATE = "state"
CREATED = "created"
UPDATED = "updated"
NO_SELECTIONS = {"__filter__": None}
CONSTANTS = [str, int, float, bool]
COLUMNS = "columns"
VIRTUAL = "virtual"
OUTPUT_TYPE = "output_type"
NOT_IMPLEMENTED_ERROR = "Not implemented"
JSON = "json"
PANDAS = "pandas"
VAEX = "vaex"
SKLEARN = "sklearn"
NUMPY = "numpy"
SELECTIONS = "selections"
SELECTIONS_FILTER = "__filter__"
VARIABLES = "variables"
DESCRIPTION = "description"
RAW = "raw"
PIPELINE = "pipeline"
PIPELINE_TYPE = "pipeline_type"
STATE = "state"
VERSION = "version"
PY_VERSION = "py_version"
PACKAGES = "packages"
VALID_VAEX_HEADERS = ["*.csv", "*.hdf5", "*.parquet", "*.arrow"]
DEFAULT_SUFFIX = ".parquet"
BYTE_DELIMITER = b'###'

# App
CORS_ORIGINS = os.getenv('ORIGINS', '*').split(',')
ALLOW_METHODS = os.getenv('ALLOW_METHODS', '*').split(',')
ALLOW_HEADERS = os.getenv('ALLOW_HEADERS', '*').split(',')
ALLOW_CREDENTIALS = bool(strtobool(os.getenv('ALLOW_CREDENTIALS', 'True')))
ALLOW_CORS = bool(strtobool(os.getenv('ALLOW_CORS', 'True')))



from setuptools import setup, find_packages

packages = find_packages()
print(packages)
setup(
    name="goldilox",
    packages=packages,
    include_package_data=True,
    install_requires=["sklearn", "pandas", "numpy", "cloudpickle", "gunicorn", "Click", "fastapi", "uvicorn"],
    version="0.0.3",
    url="https://github.com/xdssio/goldilox",
    description="A tool for deploying machine learning",
    author="Yonatan Alexander",
    author_email="jonathan@xdss.io",
    entry_points="""
        [console_scripts]
        glx=goldilox.app.cli:main
    """,
    python_requires='>=3',
)



import uuid
from collections import OrderedDict

import numpy as np
import six
from vaex.expression import Expression


def is_ndarray(value):
    return isinstance(value, np.ndarray)


def is_numpy_numeric(dtype):
    if is_expression(dtype):
        dtype = dtype.dtype
    return is_numpy_float(dtype) or is_numpy_int(dtype)


def is_numpy_datetime(dtype):
    if is_expression(dtype):
        dtype = dtype.dtype
    return hasattr(dtype, 'char') and dtype.char in np.typecodes.get('Datetime')


def is_numpy_int(dtype):
    if is_expression(dtype):
        dtype = dtype.dtype
    return hasattr(dtype, 'char') and dtype.char in np.typecodes.get('AllInteger')


def is_numpy_float(dtype):
    if is_expression(dtype):
        dtype = dtype.dtype
    return hasattr(dtype, 'char') and dtype.char in np.typecodes.get('AllFloat')


def is_hidden(s):
    return str(s).startswith('__')


def is_expression(e):
    return isinstance(e, Expression)


def unqiue_with_order(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def _ensure_string_from_expression(expression):
    import vaex.expression
    if expression is None:
        return None
    elif isinstance(expression, bool):
        return expression
    elif isinstance(expression, six.string_types):
        return expression
    elif isinstance(expression, vaex.expression.Expression):
        return expression.expression
    else:
        raise ValueError('%r is not of string or Expression type, but %r' % (expression, type(expression)))


def validate_list(obj):
    if obj is not None and not isinstance(obj, list):
        return [obj]
    return obj


class OrderedIndexedDict(OrderedDict):

    def get(self, item, defualt=None):
        ret = super(OrderedIndexedDict, self).get(item, defualt)
        if ret is None and isinstance(item, int):
            keys = list(self.keys())
            ret = super(OrderedIndexedDict, self).get(keys[item], defualt)
        return ret

    def __getitem__(self, item):
        return self.get(item)


def get_random_id():
    return str(uuid.uuid4()).split('-')[0]


def is_float(value):
    try:
        return float(value)
    except ValueError:
        return False



import numpy as np
VALUE = 'VALUE'
MODE = 'MODE'
UNIQUE = 'UNIQUE'
COMMON = 'COMMON'
NEW_VALUE = 'NEW_VALUE'
TRANSFORMER = 'TRANSFORMER'
MEAN = 'MEAN'
SUM = 'SUM'
MIN = 'MIN'
MAX = 'MAX'
STD = 'STD'
VAR = 'VAR'
SIZE = 'SIZE'
COUNT = 'COUNT'
SELF = 'SELF'
MEDIAN = 'MEDIAN'
SUMMARY = 'SUMMARY'
DTYPE = 'DTYPE'
ROLLING_MEAN = '_ROLLING_MEAN'
ROLLING_STDV = '_ROLLING_STDV'
EQ = '=='
NE = '!='
GT = '>'
LT = '<'
GE = '>='
LE = '<='
ADD = '+'
SUB = '-'
MUL = '*'
POWER = '**'
TRUE_DIV = '/'
FLOOR_DIV = '//'
AND = '&'
OR = '|'
DROP_NA = 'DROP_NA'
FILLNA = 'FILL_NA'
COUNTNA = 'COUNTNA'
NUM_UNIQUE = 'NUM_UNIQUE'
HISTOGRAM = 'HISTOGRAM'
VALUES = 'VALUES'
NON_MISSING_COUNT = 'NON_MISSING_COUNT'
NUM_ZERO = 'NUM_ZERO'
NUM_EMPTY_STRINGS = 'NUM_EMPTY_STRINGS'
VALUE_COUNTS = 'FREQUENT_ITEMS'
NONE_TYPE = type(None)
NONE_STRING = 'NONE'
EXPRESSION = '_EXPRESSION_'
COLUMN = 'COLUMN'
APPLY = 'APPLY'
FILTER = 'FILTER'
SEQUENCE = 'SEQUENCE'
LOGICAL = 'LOGICAL'
ARRAY = 'ARRAY'
CONSTANT = 'CONSTANT'
BOW = '_BAG_OF_WORDS'
COUNT_NGRAM = '_COUNT_NGRAM'
TOKENIZE = '_TOKENIZE'
DROP_WORDS = '_DROP_WORDS'
TRIM_BY_VALUES = '_TRIM_BY_VALUES'
TRIM_BY_KEYS = '_TRIM_BY_KEYS'
PACK = '_PACK'
UNPACK = '_UNPACK'
STACK = '_STACK'
UNSTACK = '_UNSTACK'
FUNCTION = 'FUNCTION'
TRANSFORMER = 'TRANSFORMER'
LOGICAL_COMPARISONS = [EQ, NE, LE, GE, LT, GT]
NUMERIC_COMPARISONS = [ADD, SUB, MUL, POWER, FLOOR_DIV, TRUE_DIV, OR, AND]
COMPARISONS = LOGICAL_COMPARISONS + NUMERIC_COMPARISONS
STATISTICS = [MEAN, SUM, VAR, STD, MIN, MAX, COUNTNA, MODE, COMMON, NEW_VALUE, ROLLING_MEAN, SIZE, NUM_UNIQUE, NUM_ZERO,
              VALUE_COUNTS, NUM_EMPTY_STRINGS]
TRANSFORMERS_OPERATORS = [BOW, COUNT_NGRAM, TRIM_BY_VALUES, TRIM_BY_KEYS, DROP_NA, TOKENIZE, DROP_WORDS]
OPERATORS = COMPARISONS + STATISTICS + TRANSFORMERS_OPERATORS + [UNIQUE]
AUTO = 'auto'
COLLABORATIVE_FILTERING = 'COLLABORATIVE_FILTERING'
MATRIX_FACTORIZATION = 'MATRIX_FACTORIZATION'
RANKING_FACTORIZATION = 'RANKING_FACTORIZATION'
CONTENT_RECOMMENDER = 'CONTENT_RECOMMENDER'
POPULARITY_RECOMMENDER = 'POPULARITY_RECOMMENDER'

RECOMMENDERS = [CONTENT_RECOMMENDER, COLLABORATIVE_FILTERING, MATRIX_FACTORIZATION, RANKING_FACTORIZATION,
                POPULARITY_RECOMMENDER, AUTO]

RECOMMENDATIONS_MODULES = ['user', 'provider', 'product']

# config strings
DATAFRAME_CONFIG_FILENAME = 'dataframe.ini'
PIPELINE = 'pipeline'
DTYPE = 'dtype'
DATAFRAME = 'dataframe'
UID = 'uid'
CHAIN = 'chain'
LOCATION = 'location'
VALUES = 'values'
FEATURES = 'features'
EXPRESSIONS = 'expressions'
NAME = 'name'
STATE = 'state'
INPLACE = 'inplace'
VERBOSE = 'verbose'
OUTPUT_COLUMN = 'output_column'
OPTIONS = 'options'
COLUMNS = 'columns'
STEPS = 'steps'
CLASSIFICATION = 'classification'
REGRESSION = 'regression'
TARGET = 'target'
RANKING = 'ranking'
PARAMS = 'params'
FN = 'fn'
DS = 'ds'
marshal_encoding = 'ISO-8859-1'
MODEL = 'model'
PREFIX = 'prefix'
CACHE = 'cache'
__PROXY__ = '__proxy__'
TRAINING_TIME = 'training_time'
CLASS = 'class'
CLASSIFIR_CLASS = 'class'
CLASSIFIR_PROBABILITIS = 'probability'
CLASSIFIR_MARGIN = 'margin'
CLASSIFIR_VECTOR = 'probability_vector'
TOPK = 'topk'
TOPK_PROBABILITY = 'topk_probability'
TOPK_MARGIN = 'topk_margin'
TOPK_RANK = 'topk_rank'
COLUMNS = 'columns'
VARIABLES = 'variables'
SELECTIONS = 'selections'
STATE = 'state'
SELECTIONS_FILTER = '__filter__'
NOT_IMPLEMENTED_ERROR = 'Not implemented'
OUTPUT_TYPE = 'output_type'
JSON = 'json'
PANDAS = 'pandas'
VAEX = 'vaex'
NUMPY = 'numpy'
VIRTUAL = 'virtual'
EXAMPLE = 'example'
CREATED = 'created'
UPDATED = 'updated'
NO_SELECTIONS = {'__filter__': None}
CONSTANTS = [str, int, float, bool]
LISTS = 'LISTS_'
PACK = 'PACK_'
UNPACK = 'UNPACK_'
DATETIME_DTYPE = np.dtype('<M8[ns]')


import logging
from typing import List

from pydantic import create_model

from goldilox import Pipeline
from goldilox.utils import to_nulls, process_variables

PIPELINE = "pipeline"
RAW = "raw"
GUNICORN = "gunicorn"
UVICORN = "uvicorn"
PATH = "path"


def parse_query(query):
    if not isinstance(query, list):
        query = [query]
    return [q.dict() for q in query if q]


def process_response(items):
    items = Pipeline.to_records(items)
    if not isinstance(items, list):
        items = [items]
    for item in items:
        for key, value in item.items():
            item[key] = to_nulls(value)
    return items


def get_app(path):
    from fastapi import FastAPI, HTTPException
    from goldilox.config import ALLOW_CORS, CORS_ORIGINS, ALLOW_HEADERS, ALLOW_METHODS, ALLOW_CREDENTIALS
    logger = logging.getLogger(__name__)
    app = FastAPI()
    if ALLOW_CORS:
        from fastapi.middleware.cors import CORSMiddleware
        app.add_middleware(
            CORSMiddleware,
            allow_origins=CORS_ORIGINS,
            allow_credentials=ALLOW_CREDENTIALS,
            allow_methods=ALLOW_METHODS,
            allow_headers=ALLOW_HEADERS,
        )
    PIPELINE = "pipeline"
    RAW = "raw"

    pipeline = Pipeline.from_file(path)
    # A dynamic way to create a pydanic model based on the raw data
    raw = process_response(pipeline.raw)[0]
    Query = create_model(
        "Query",
        **{k: (type(v), None) for k, v in raw.items()},
        __config__=type("QueryConfig", (object,), {"schema_extra": {"example": raw}}),
    )

    def get_raw():
        return app.state._state.get(RAW, {})

    def get_pipeline():
        return app.state._state.get(PIPELINE, pipeline)

    @app.post("/inference", response_model=List[dict])
    def inference(data: List[Query], columns: str = ""):
        logger.info("/inference")
        data = parse_query(data)
        if len(data) == 0:
            raise HTTPException(status_code=400, detail="No data provided")
        try:
            columns = None if not columns else columns.split(",")
            ret = get_pipeline().inference(data, columns=columns)

        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=str(
                f"Issue with inference, try runing your pipeline locally with 'pipeline.inference(data)' to see what is the problem:\n{str(e)}"))

        return process_response(ret)

    @app.get("/variables", response_model=dict)
    def variables():
        logger.info("/variables")
        return process_variables(get_pipeline().variables)

    @app.get("/description", response_model=str)
    def description():
        logger.info("/description")
        return get_pipeline().description

    @app.get("/example", response_model=List[dict])
    def example():
        logger.info("/example")
        return process_response(get_pipeline().example)

    return app


def get_wsgi_application(path):
    from gunicorn.app.base import BaseApplication

    class WSGIApplication(BaseApplication):
        def __init__(self, app, options=None):
            self.options = options or {}
            self.application = app
            super().__init__()

        # # per worker
        def load_config(self):
            config = {
                key: value
                for key, value in self.options.items()
                if key in self.cfg.settings and value is not None
            }
            for key, value in config.items():
                self.cfg.set(key.lower(), value)
            pipeline = Pipeline.from_file(path)
            self.application.state._state[PIPELINE] = pipeline
            self.application.state._state[RAW] = pipeline.raw.copy()

        def load(self):
            return self.application

    return WSGIApplication


class Server:
    def __init__(self, path, options={}):
        self.path = path
        self.options = options
        self.app = get_app(path)
        pipeline = Pipeline.from_file(path)
        if not pipeline.validate():
            raise RuntimeError(f"Pipeline in {path} is invalid")

    @staticmethod
    def _validate_worker_class(options):
        options["worker_class"] = options.get(
            "worker_class", "uvicorn.workers.UvicornH11Worker"
        )
        return options

    def serve(self, options=None):
        options = self._validate_worker_class(options or self.options)
        WSGIApplication = get_wsgi_application(self.path)

        WSGIApplication(self.app, options).run()



import inspect
import json
import logging
from collections import OrderedDict
from copy import deepcopy, copy as _copy
from glob import glob
from numbers import Number
from time import time

import cloudpickle
import numpy as np
import pandas as pd
import pyarrow as pa
import traitlets
import vaex
from vaex.column import Column
from vaex.ml.state import HasState, serialize_pickle

from goldilox.config import *
from goldilox.pipeline import Pipeline
from goldilox.utils import process_variables

logging.basicConfig()
logger = logging.getLogger(__name__)

PIPELINE_FIT = "_PIPELINE_FIT"
FUNCTIONS = "functions"
VARIABLES = "variables"


class VaexPipeline(HasState, Pipeline):
    pipeline_type = "vaex"
    current_time = int(time())
    created = traitlets.Int(
        default_value=current_time, allow_none=False, help="Created time"
    )
    updated = traitlets.Int(
        default_value=current_time, allow_none=False, help="Updated time"
    )
    raw = traitlets.Any(
        default_value=None, allow_none=True, help="An example of the raw dataset"
    ).tag(**serialize_pickle)
    _original_columns = traitlets.List(
        default_value=[], help="original columns which were not virtual expressions"
    )
    state = traitlets.Dict(
        default_value=None, allow_none=True, help="The state to apply on inference"
    )
    description = traitlets.Unicode(
        default_value="", help="Any notes to associate with a pipeline instance"
    )
    variables = traitlets.Dict(
        default_value={}, help="Any variables to associate with a pipeline instance"
    )

    @property
    def example(self):
        return self.inference(self.raw).to_records(0)

    @classmethod
    def _get_original_columns(cls, df):
        return list(df.dataset._columns.keys())

    @classmethod
    def _data_type(cls, data):

        if isinstance(data, np.ndarray):
            data_type = data.dtype
        elif isinstance(data, Column):
            data = pa.array(data.tolist())
            data_type = data.type.to_pandas_dtype()
        else:
            # when we eval constants, let arrow find it out

            if isinstance(data, Number):
                data_type = pa.array([data]).type.to_pandas_dtype()
            else:
                data_type = data.type.to_pandas_dtype()  # assuming arrow
        return data_type

    @classmethod
    def _get_original_dtypes(cls, df):
        return {
            column: cls._data_type(data)
            for column, data in df.head(1).dataset._columns.items()
        }

    def state_set(self, state):
        HasState.state_set(self, state)
        self.updated = int(time())
        return self

    def not_implemented(self):
        return None

    @property
    def virtual_columns(self):
        return self.state.get("virtual_columns")

    @property
    def functions(self):
        return self.state.get("functions")

    @classmethod
    def verify_vaex_dataset(cls, df):
        if not cls.is_vaex_dataset(df):
            raise ValueError(
                "ds should be a vaex.dataset.DatasetArrays or vaex.hdf5.dataset.Hdf5MemoryMapped"
            )
        return df.copy()

    @classmethod
    def from_dict(cls, state):
        if "state" in state:
            ret = VaexPipeline()
            ret.state_set(state)
            return ret
        return VaexPipeline(state=state, example=None, fit_func=None)

    def sample_first(self, df):
        if len(df) == 0:
            raise RuntimeError("cannot sample from empty dataframe")
        try:
            sample = df[0:1]
            self._original_columns = self._get_original_columns(df)
            self.raw = {
                key: values[0] for key, values in sample.dataset._columns.items()
            }
            return True
        except Exception as e:
            logger.error(f"could not sample first: {e}")
        return False

    @classmethod
    def _tolist(cls, value):
        if hasattr(value, "ar"):
            return value.ar.tolist()
        return value.tolist()

    @classmethod
    def from_dataframe(cls, df, fit=None, variables=None, description=""):
        """
       Get a Pipeline out of a vaex.dataframe.DataFrame, and validate serilization and missing values.
       @param df: vaex.dataframe.DataFrame
       @param fit: method: A method which accepts a vaex dataframe and returns a vaex dataframe which run on pipeline.fit(df).
       @param variables: dict [optional]: Any variables we want to associate with the current pipeline.
              On top of the variables provided, the dataframe variables are added.
       @param description: str [optional]: Any text we want to associate with the current pipeline.
       @return: VaexPipeline
       """
        copy = VaexPipeline.verify_vaex_dataset(df)
        if fit is not None:
            copy.add_function(PIPELINE_FIT, fit)
        sample = copy[0:1]
        state = copy.state_get()
        renamed = {x[1]: x[0] for x in state["renamed_columns"]}
        raw = {
            renamed.get(key, key): cls._tolist(value)[0]
            for key, value in sample.dataset._columns.items()
        }
        original_columns = VaexPipeline._get_original_columns(sample)
        variables = {**(state.get(VARIABLES) or {}), **(variables or {})}

        pipeline = VaexPipeline(
            state=state,
            _original_columns=original_columns,
            raw=raw,
            description=description,
            variables=process_variables(variables),
        )

        return pipeline

    def copy(self):
        pipeline = VaexPipeline(state={})
        pipeline.state_set(deepcopy(self.state_get()))
        pipeline.updated = int(time())
        return pipeline

    @classmethod
    def from_file(cls, path):
        return Pipeline.from_file(path)

    def _dumps(self):
        return cloudpickle.dumps(_copy(self.state_get()))

    @classmethod
    def json_load(cls, state):
        from vaex.json import VaexJsonDecoder

        return json.loads(state[STATE], cls=VaexJsonDecoder)

    @classmethod
    def is_vaex_dataset(cls, ds):
        return isinstance(ds, vaex.dataframe.DataFrame)

    @classmethod
    def load_state(cls, state):
        instance = VaexPipeline()
        if not isinstance(state, dict):
            state = VaexPipeline.json_load(state)
        instance.state_set(state)
        return instance

    def __getitem__(self, index):
        virtual_columns = list(OrderedDict(self.virtual_columns).items())
        return virtual_columns[index]

    # TODO replace add_memory column - currently not working
    def add_virtual_column(self, df, name, value, first_column):
        df[name] = df.func.where(df[first_column] == value, value, value)
        return df

    def add_memmory_column(self, df, name, value, length=None):
        if length is None:
            length = len(df)
        df[name] = np.array([value] * length)
        return df

    def na_column(self, length):
        return pa.array([None] * length)

    def preprocess_transform(
            self,
            df,
    ):
        copy = self.infer(df)
        length = len(copy)

        renamed = {x[1]: x[0] for x in self.state["renamed_columns"]}
        for key in self._original_columns:
            if key in renamed:
                key = renamed.get(key)
            if key not in copy:
                copy[key] = self.na_column(length)
        return copy

    def transform_state(self, df, keep_columns=None, state=None, set_filter=False):
        copy = df.copy()
        state = state or self.state
        if state is not None:
            if keep_columns is True:
                keep_columns = list(
                    set(copy.get_column_names()).difference(
                        [k for k in state["column_names"] if not k.startswith("__")]
                    )
                )
            if keep_columns is False or (
                    keep_columns is not None and len(keep_columns) == 0
            ):
                keep_columns = None
            copy.state_set(state, keep_columns=keep_columns, set_filter=set_filter)
        return copy

    def transform(self, df, keep_columns=False, state=None, set_filter=True):
        copy = self.preprocess_transform(df)
        copy = self.transform_state(
            copy, keep_columns=keep_columns, state=state, set_filter=set_filter
        )
        return copy

    def fill_columns(self, df, columns=None, length=None):
        if columns is None:
            return df
        if length is None:
            length = len(df)
        for column in columns:
            if column not in df:
                value = self.values.get(column)
                if value is not None:
                    self.add_memmory_column(df, column, value, length)
        return df

    def inference(
            self, df, columns=None, passthrough=True, set_filter=False,
    ):
        if isinstance(columns, str):
            columns = [columns]
        copy = self.preprocess_transform(df)
        ret = self.transform_state(
            copy, set_filter=set_filter, keep_columns=passthrough
        )
        # ret = self.fill_columns(copy, columns=columns)
        if columns is not None:
            ret = ret[columns]
        return ret

    def evaluate(self):
        raise NotImplementedError("evaluate not implemented")

    def partial_fit(self, start_index=None, end_index=None):
        raise ValueError("partial_fit implemented")

    @classmethod
    def infer(cls, data, **kwargs):
        if isinstance(data, vaex.dataframe.DataFrame):
            return data.copy()
        elif isinstance(data, pd.DataFrame):
            return vaex.from_pandas(data)
        elif isinstance(data, str):
            if os.path.isfile(data):
                logger.info(f"reading file from {data}")
                return vaex.open(data, **kwargs)
            elif os.path.isdir(data):
                logger.info(f"reading files from {data}")
                files = []
                for header in VALID_VAEX_HEADERS:
                    files.extend(glob(f"{data}/{header}"))
                logger.info(f"open files: {files}")
                bad_files = set([])
                for file in files:
                    try:
                        if file.endswith(".csv"):
                            temp = vaex.open(file, nrows=1)
                        else:
                            temp = vaex.open(file)
                        temp.head()
                    except:
                        bad_files.add(file)
                files = [file for file in files if file not in bad_files]
                return vaex.open_many(files, **kwargs)
            data = json.loads(data)
        elif isinstance(data, bytes):
            data = json.loads(data)
        if isinstance(data, np.ndarray):
            columns = kwargs.get("names")
            if columns is None:
                raise RuntimeError(
                    "can't infer numpy array without 'names' as a list of columns"
                )
            if len(columns) == data.shape[1]:
                data = data.T
            return vaex.from_dict({key: value for key, value in zip(columns, data)})
        elif isinstance(data, list):
            # try records
            return vaex.from_pandas(pd.DataFrame(data))
        elif isinstance(data, dict):
            sizes = [
                0
                if (not hasattr(value, "__len__") or isinstance(value, str))
                else len(value)
                for value in data.values()
            ]
            if len(sizes) == 0 or min(sizes) == 0:
                data = data.copy()
                for key, value in data.items():
                    if isinstance(value, list) or isinstance(value, np.ndarray):
                        data[key] = np.array([value])
                    else:
                        data[key] = [value]
            random_value = data[list(data.keys())[0]]
            if isinstance(random_value, dict):
                return vaex.from_pandas(pd.DataFrame(data))
            return vaex.from_arrays(**data)

        raise RuntimeError("Could not infer a vaex type")

    def set_variable(self, key, value):
        self.variables[key] = value
        return value

    def get_variable(self, key, default=None):
        return self.variables.get(key, default)

    def get_columns_names(self, virtual=True, strings=True, hidden=False, regex=None):
        return self.inference(self.raw).get_column_names(
            virtual=virtual, strings=strings, hidden=hidden, regex=regex
        )

    def fit_transform(self, data):
        self.fit(data)
        return self.transform(data)

    def is_valid_fit(self, f):
        return f is not None and callable(f) and len(inspect.getfullargspec(f)[0]) > 0

    def set_fit(self, f=None):
        if self.is_valid_fit(f):
            self.fit_func = f

    def get_function(self, name):
        from vaex.serialize import from_dict

        return from_dict(self.state.get(FUNCTIONS, {}).get(name), trusted=True).f

    def fit(self, df):
        copy = df.copy()
        self.verify_vaex_dataset(copy)
        self.sample_first(copy)
        fit_func = self.get_function(PIPELINE_FIT)
        if fit_func is None:
            raise RuntimeError("'fit()' was not set for this pipeline")
        self.raw = copy.to_records(0)
        trained = fit_func(copy)
        if VaexPipeline.is_vaex_dataset(trained):
            trained.add_function(PIPELINE_FIT, fit_func)
            self.state = trained.state_get()
        else:
            if isinstance(trained, dict):
                self.state = trained
            else:
                raise ValueError(
                    "'fit_func' should return a vaex dataset or a state, got {}".format(
                        type(trained)
                    )
                )
        self.variables.update(self.state.get(VARIABLES, {}))
        self.updated = int(time())
        return self

    def get_function_model(self, name):
        tmp = self.state["functions"][name]
        model = eval(tmp["cls"])()
        model.state_set(state=tmp["state"])
        return model



import logging
import os
from functools import partial
from pathlib import Path

import PIL.Image
import ipyplot
import numpy as np
import tensorflow as tf
import tensorflow_io.arrow as arrow_io
import vaex
from tensorflow.python.keras.models import model_from_json
from tensorflow_io.arrow.python.ops.arrow_dataset_ops import arrow_schema_to_tensor_types
from vaex.ml.state import HasState

logger = logging.getLogger("keras")


class KerasModel(HasState):

    def __init__(self, model=None):
        try:
            model = model_from_json(model)
        except:
            pass
        self.model = model

    def __reduce__(self):
        return (self.__class__, (self.model.to_json(),))

    def predict(self, ar):
        if isinstance(ar, vaex.expression.Expression):
            ar = ar.values
        return self.model.predict(ar)


@vaex.register_dataframe_accessor('keras', override=True)
class DataFrameAccessorTensorflow(object):
    def __init__(self, df):
        self.df = df

    def _arrow_batch_generator(self, column_names, chunk_size=1024, parallel=True):
        """Create a generator which yields arrow table batches, to use as datasoure for creating Tensorflow datasets.
        :param features: A list of column names.
        :param target: The dependent or target column, if any.
        :param chunk_size: Number of samples per chunk of data.
        :param parallel: If True, vaex will evaluate the data chunks in parallel.
        Returns:
        :return: generator that yields arrow table batches
        """
        for i1, i2, table in self.df.to_arrow_table(column_names=column_names, chunk_size=chunk_size,
                                                    parallel=parallel):
            yield table.to_batches(chunk_size)[0]

    @staticmethod
    def _get_batch_arrow_schema(arrow_batch):
        """Get the schema from a arrow batch table."""
        output_types, output_shapes = arrow_schema_to_tensor_types(arrow_batch.schema)
        return output_types, output_shapes

    def to_dataset(self, features, target=None, chunk_size=1024, as_dict=True, parallel=True):
        """Create a tensorflow Dataset object from a DataFrame, via Arrow.
        :param features: A list of column names.
        :param target: The dependent or target column, if any.
        :param chunk_size: Number of samples per chunk of data.
        :param as_dict: If True, the dataset will have the form of dictionary housing the tensors.
        :param parallel: If True, vaex will evaluate the data chunks in parallel.
        This is useful for making inputs directly for tensorflow. If False, the dataset will contain Tensors,
        useful for passing the dataset as a datasource to a Keras model.
        Returns:
        :return ds: A tensorflow Dataset
        """
        if target is not None:
            target = vaex.utils._ensure_list(target)
            target = vaex.utils._ensure_strings_from_expressions(target)
            n_target_cols = len(target)
            column_names = features + target
        else:
            column_names = features

        # Set up the iterator factory
        iterator_factory = partial(self._arrow_batch_generator, **{'column_names': column_names,
                                                                   'chunk_size': chunk_size,
                                                                   'parallel': parallel})
        # get the arrow schema
        output_types, output_shapes = self._get_batch_arrow_schema(next(iterator_factory()))

        # Define the TF dataset
        ds = arrow_io.ArrowStreamDataset.from_record_batches(record_batch_iter=iterator_factory(),
                                                             output_types=output_types,
                                                             output_shapes=output_shapes,
                                                             batch_mode='auto',
                                                             record_batch_iter_factory=iterator_factory)

        # Reshape the data into the appropriate format
        if as_dict:
            if target is not None:
                if n_target_cols == 1:
                    ds = ds.map(lambda *tensors: (dict(zip(features, tensors[:-1])), tensors[-1]))
                else:
                    ds = ds.map(lambda *tensors: (dict(zip(features, tensors[:-n_target_cols])),
                                                  # dict(zip(target, tensors[-n_target_cols:]))))
                                                  tf.stack(tensors[-n_target_cols:], axis=1)))
            else:
                ds = ds.map(lambda *tensors: (dict(zip(features, tensors))))
        else:
            if target is not None:
                if n_target_cols == 1:
                    ds = ds.map(lambda *tensors: (tf.stack(tensors[:-1], axis=1), tensors[-1]))
                else:
                    ds = ds.map(lambda *tensors: (tf.stack(tensors[:-n_target_cols], axis=1),
                                                  tf.stack(tensors[-n_target_cols:], axis=1)))
            else:
                ds = ds.map(lambda *tensors: (tf.stack(tensors, axis=1)))

        return ds

    def make_input_function(self, features, target=None, chunk_size=1024, repeat=None, shuffle=False, parallel=True):
        """Create a tensorflow Dataset object from a DataFrame, via Arrow.
        :param features: A list of column names.
        :param target: The dependent or target column, if any.
        :param chunk_size: Number of samples per chunk of data.
        :param repeat: If not None, repeat the dataset as many times as specified.
        :param shuffle: If True, the elements of the dataset are randomly shuffled. If shuffle is True and repeat is not None,
        the dataset will first be repeated, and the entire repeated dataset shuffled.
        :param parallel: If True, vaex will evaluate the data chunks in parallel.
        Returns:
        :return ds: A tensorflow Dataset
        """
        if repeat is not None:
            assert (isinstance(repeat, int)) & (
                    repeat > 0), 'The "repeat" arg must be a positive integer larger larger than 0.'
            shuffle_buffer_size = chunk_size * repeat
        else:
            shuffle_buffer_size = chunk_size

        def tf_input_function():
            ds = self.to_dataset(features=features, target=target, chunk_size=chunk_size, parallel=True)
            if repeat is not None:
                ds = ds.repeat(repeat)
            if shuffle:
                ds = ds.shuffle(shuffle_buffer_size)

            return ds

        return tf_input_function

    def to_keras_generator(self, features, target=None, chunk_size=1024, parallel=True, verbose=True):
        """Return a batch generator suitable as a Keras datasource.  Note that the generator is infinite, i.e. it loops
        continuously ovder the data. Thus you need to specify the "steps_per_epoch" arg when fitting a keras model,
        the "validation_steps" when using it for validation, and "steps" when calling the "predict" method of a keras model.
        :param features: A list of column names.
        :param target: The dependent or target column or a list of columns, if any.
        :param chunk_size: Number of samples per chunk of data. This can be thought of as the batch size.
        :param parallel: If True, vaex will evaluate the data chunks in parallel.
        :parallel verbose: If True, show an info on the recommended "steps_per_epoch"
        based on the total number of samples and "chunk_size".
        """
        if verbose:
            steps_per_epoch = np.ceil(len(self.df) / chunk_size)
            logging.info(f'Recommended "steps_per_epoch" arg: {steps_per_epoch}')

        target = target

        def _generator(features, target, chunk_size, parallel):
            if target is not None:
                target = vaex.utils._ensure_list(target)
                target = vaex.utils._ensure_strings_from_expressions(target)
                n_target_cols = len(target)
                column_names = features + target

            else:
                column_names = features
            column_shapes = [self.df[column].shape for column in column_names]
            number_of_features = len(features)

            while True:
                if target is not None:
                    for i1, i2, chunks in self.df.evaluate_iterator(column_names, chunk_size=chunk_size,
                                                                    parallel=parallel):
                        chunks = [np.array(feature_data) for feature_data in chunks]
                        if len(chunks[0]) < chunk_size:
                            return
                        X = chunks[:number_of_features]
                        if len(X) == 1:
                            X = X[0]
                        else:
                            raise NotImplementedError("need to do")
                        y = chunks[-1]
                        yield (X, y)
                else:
                    for i1, i2, chunks, in self.df.evaluate_iterator(column_names, chunk_size=chunk_size,
                                                                     parallel=parallel):
                        X = np.array(chunks).T
                        yield (X,)

        return _generator(features, target, chunk_size, parallel)

    def persists_model_predict(self, model):
        return KerasModel(model).predict

    def show(self, x, labels=None, max_images=20, img_width=85):
        x = vaex.utils._ensure_list(x)[0]
        images = self.df[x].values
        if labels is not None:
            labels = vaex.utils._ensure_list(labels)[0]
            labels = self.df[labels].values
        else:
            labels = [''] * len(images)
        ipyplot.plot_images(images, max_images=max_images, img_width=img_width, labels=labels)
        return None


def open_images(path, suffix=None, resize=None):
    if os.path.isfile(path):
        files = [path]
    if os.path.isdir(path):
        files = []
        if suffix is not None:
            files = [str(path) for path in Path(path).rglob(f"*{suffix}")]
        else:
            for suffix in ['jpg', 'png', 'jpeg', 'ppm', 'thumbnail']:
                files.extend([str(path) for path in Path(path).rglob(f"*{suffix}")])
    num_skipped = 0
    ignores = set([])
    for file in files:
        try:
            fobj = open(file, "rb")
            is_jfif = tf.compat.as_bytes("JFIF") in fobj.peek(10)
        finally:
            fobj.close()
        if not is_jfif:
            num_skipped += 1
            logger.error(f"file {path} is corrupted - ignore")
            ignores.add(file)
    files = [file for file in files if file not in ignores]
    df = vaex.from_arrays(path=files)
    logger.info(f"found {len(files)} files")

    # if shape is None:
    #     shape = np.array(PIL.Image.open(files[0])).shape
    #
    # def read_image(path):
    #     try:
    #         im = PIL.Image.open(path)
    #         im = resize(im, shape)
    #         return np.array(im)
    #     except:
    #         logger.error(f"failed to open - {path}")
    #     return None
    @vaex.register_function()
    def imopen(paths):
        images = [PIL.Image.open(path) for path in vaex.array_types.tolist(paths)]
        return np.array(images, dtype="O")

    df['image'] = df['path'].imopen()

    return df



import numpy as np
import pandas as pd
import vaex
import json

@vaex.register_dataframe_accessor('gl', override=True)
class GoldiloxExternsion(object):
    def __init__(self, df):
        self.df = df

    def countna(self):
        columns = self.df.get_column_names()
        return sum([self.df[feature].countna() for feature in columns])

    @staticmethod
    def to_python(value):
        if isinstance(value, np.ndarray):
            value = value.tolist()
        return value

    def to_pandas(self, chunk_size=None):
        if chunk_size is None:
            return pd.DataFrame.from_dict({key: list(value) for key, value in self.df.to_items()})

        def iterator():
            for _, _, chunk in self.df.to_items(chunk_size=chunk_size):
                yield pd.DataFrame.from_dict({key: list(value) for key, value in chunk})

        return iterator()

    def to_records(self, item=None, chunk_size=None):
        if isinstance(item, int):
            return {key: value[0] for key, value in self.df[item:item + 1].to_dict(array_type='python').items()}
            # return {key: self.to_python(value) for key, value in zip(self.df.get_column_names(), self.df[item])}
        if item is not None:
            raise RuntimeError(f"item can be None or an int - {type(item)} provided")
        else:
            if chunk_size is None:
                records = self.df.to_dict(array_type='python')
                keys = list(records.keys())
                return [{key:value for key,value in zip(keys,values)} for values in zip(*records.values())]

            def iterator():
                for _,_ ,chunk in self.df.to_dict(chunk_size=chunk_size, array_type='python'):
                    keys = list(chunk.keys())
                    yield [{key:value for key,value in zip(keys,values)} for values in zip(*chunk.values())]

            return iterator()

    def json_dumps(self, item=None):
        return json.dumps(self.to_records(item=item))

    def json_dump(self, fp, item=None):
        return json.dump(self.to_records(item=item), fp=fp)

